/* 
Copyright Paul James Mutton, 2001-2004, http://www.jibble.org/

This file is part of PieSpy.

This software is dual-licensed, allowing you to choose between the GNU
General Public License (GPL) and the www.jibble.org Commercial License.
Since the GPL may be too restrictive for use in a proprietary application,
a commercial license is also provided. Full license information can be
found at http://www.jibble.org/licenses/

$Author: pjm2 $
$Id: Graph.java,v 1.38 2004/05/11 15:07:04 pjm2 Exp $

*/

package org.jibble.socnet;

import java.util.*;
import java.io.*;
import java.awt.*;
import java.awt.image.*;
import java.text.*;
import javax.imageio.ImageIO;

/**
 * The Graph stores the Nodes and Edges, and InferenceHeurisics to allow
 * the structure of the graph to be modified.
 */
public class Graph implements java.io.Serializable {
    static final long serialVersionUID = 1L;
    String title = "A Social Network Diagram for an IRC Channel";
    String desc1 = "";
    String desc2 = "Blue edge thickness and shortness represents strength of relationship";
    String desc3 = null;
    String[] box_alpha = { "A0", "A1", "A2", "A3", "A4", "A5", "A6", "A7", "A8", "A9", "B0", "B1", "B2", "B3", "B4", "B5", "B6", "B7", "B8", "B9", "C0", "C1", "C2", "C3", "C4", "C5", "C6", "C7", "C8", "C9", "D0", "D1", "D2", "D3", "D4", "D5", "D6", "D7", "D8", "D9", "E0", "E1", "E2", "E3", "E4", "E5", "E6", "E7", "E8", "E9", "F0", "F1", "F2", "F3", "F4", "F5", "F6", "F7", "F8", "F9", "G0", "G1", "G2", "G3", "G4", "G5", "G6", "G7", "G8", "G9", "H0", "H1", "H2", "H3", "H4", "H5", "H6", "H7", "H8", "H9", "I0", "I1", "I2", "I3", "I4", "I5", "I6", "I7", "I8", "I9", "J0", "J1", "J2", "J3", "J4", "J5", "J6", "J7", "J8", "J9", "K0", "K1", "K2", "K3", "K4", "K5", "K6", "K7", "K8", "K9", "L0", "L1", "L2", "L3", "L4", "L5", "L6", "L7", "L8", "L9", "M0", "M1", "M2", "M3", "M4", "M5", "M6", "M7", "M8", "M9", "N0", "N1", "N2", "N3", "N4", "N5", "N6", "N7", "N8", "N9", "O0", "O1", "O2", "O3", "O4", "O5", "O6", "O7", "O8", "O9", "P0", "P1", "P2", "P3", "P4", "P5", "P6", "P7", "P8", "P9", "Q0", "Q1", "Q2", "Q3", "Q4", "Q5", "Q6", "Q7", "Q8", "Q9", "R0", "R1", "R2", "R3", "R4", "R5", "R6", "R7", "R8", "R9", "S0", "S1", "S2", "S3", "S4", "S5", "S6", "S7", "S8", "S9", "T0", "T1", "T2", "T3", "T4", "T5", "T6", "T7", "T8", "T9", "U0", "U1", "U2", "U3", "U4", "U5", "U6", "U7", "U8", "U9", "V0", "V1", "V2", "V3", "V4", "V5", "V6", "V7", "V8", "V9", "W0", "W1", "W2", "W3", "W4", "W5", "W6", "W7", "W8", "W9", "X0", "X1", "X2", "X3", "X4", "X5", "X6", "X7", "X8", "X9", "Y0", "Y1", "Y2", "Y3", "Y4", "Y5", "Y6", "Y7", "Y8", "Y9", "Z0", "Z1", "Z2", "Z3", "Z4", "Z5", "Z6", "Z7", "Z8", "Z9" };



    public Graph(String label, Configuration config) {
        _label = label;
        this.config = config;

        // The NumberFormat used to format frame numbers in filenames.
        _nf = NumberFormat.getIntegerInstance();
        _nf.setMinimumIntegerDigits(8);
        _nf.setGroupingUsed(false);
        
        // Add InferenceHeuristics. These are processed in this order.
        _heuristics.add(new DirectAddressingInferenceHeuristic(this, config));
        _heuristics.add(new IndirectAddressingInferenceHeuristic(this, config));
        _heuristics.add(new AdjacencyInferenceHeuristic(this, config));
        _heuristics.add(new BinarySequenceInferenceHeuristic(this, config));

    }

    public void setConfiguration(Configuration config) {
	this.config = config;
    }

    // Pass a message through the list of InferenceHeuristics.
    public void infer(String nick, String message) {
        if (config.ignoreSet.contains(nick.toLowerCase())) {
            return;
        }
        
        for (int i = 0; i < _heuristics.size(); i++) {
            InferenceHeuristic heuristic = (InferenceHeuristic) _heuristics.get(i);
            heuristic.infer(nick, message);
        }
    }

    public void setTitle(String title) {
	this.title = title;
    }

    public void setDescription1(String s) {
	desc1 = s;
    }

    public void setDescription2(String s) {
	desc2 = s;
    }

    public void setDescription3(String s) {
	desc3 = s;
    }
    
    // Add a Node to the Graph.
    public void addNode(Node node) {
        
        // Only add the Node to the HashMap if it's not already in there.
        if (_nodes.containsKey(node)) {
            node = (Node) _nodes.get(node);
        }
        else {
            _nodes.put(node, node);
        }
        
        // Increment the weight of the Node.
        node.setWeight(node.getWeight() + 1);
    }

    // Add an Edge to the Graph. Increment the weighting if it already exists.    
    public boolean addEdge(Node source, Node target, double weight) {
        // Do not add self-edges or weights that are not positive.
        if (source.equals(target) || weight <= 0) {
            return false;
        }
        
        // Ensure both Nodes are in the Graph first.
        addNode(source);
        addNode(target);

        // Add the Edge to the HashMap, or find the existing entry.
        Edge edge = new Edge(source, target);
        if (_edges.containsKey(edge)) {
            edge = (Edge) _edges.get(edge);
        }
        else {
            source = (Node) _nodes.get(source);
            target = (Node) _nodes.get(target);
            edge = new Edge(source, target);
            _edges.put(edge, edge);
        }
        // Increment the edge weight.
        edge.setWeight(edge.getWeight() + weight);
        
        // The graph has changed.
        if (autoSaveEachFrame()) makeNextImage();
        return true;
    }

    boolean autoSaveEachFrame() {
	return !config.getBoolean("DontSaveEachFrame");
    }
    
    // Remove a Node from the Graph, along with all of its emanating Edges.
    public boolean removeNode(Node node) {
        if (_nodes.containsKey(node)) {
            // Remove the Node from the HashMap.
            _nodes.remove(node);
            
            // Remove all Edges that connect to the removed Node.
            Iterator edgeIt = _edges.keySet().iterator();
            while (edgeIt.hasNext()) {
                Edge edge = (Edge) edgeIt.next();
                if (edge.getSource().equals(node) || edge.getTarget().equals(node)) {
                    edgeIt.remove();
                }
            }
            return true;
        }
        return false;
    }

    private Set getNeighbours(Node node) {
	Set n = new HashSet();
	for (Iterator i = _edges.keySet().iterator();i.hasNext();) {
	    Edge edge = (Edge) i.next();
	    if (edge.getSource().equals(node))
		n.add(edge.getTarget());
	    else if (edge.getTarget().equals(node))
		n.add(edge.getSource());
	}
	return n;
    }

    public void purgeLeafs() {
	Set remove = new HashSet();
	for (Iterator i = _nodes.keySet().iterator();i.hasNext();) {
	    Node n = (Node) i.next();
	    if (getNeighbours(n).size() < 2) remove.add(n);
	}
	for (Iterator i = remove.iterator(); i.hasNext();) {
	    removeNode((Node) i.next());
	}
    }
    
    // Merges oldNode and newNode by:
    // 1) Removing newNode from the Graph and removing all emanating Edges.
    // 2) Renaming the nick of oldNode to match the nick of newNode.
    public void mergeNode(Node oldNode, Node newNode) {
        if (!this.contains(oldNode)) {
            // If the oldNode is not in this Graph, we needn't do anything.
            return;
        }
        
        // Get the nick that we want to change to.
        String nick = newNode.toString();
        
        // Remove the newNode from the Graph (and any emanating Edges).
        // Warning: Some nick changes only differ in case, so make sure
        // we're not inadvertantly removing oldNick, too!
        if (!newNode.equals(oldNode)) {
            removeNode(newNode);
        }

        // Important: HashMap keys are allocated to locations based on their
        // hashCode values when they are added. Such keys should ideally be
        // immutable, as they are assumed to return the same hashCode every
        // time.
        // Changing the nick of a Node object causes this contract to be
        // broken, which allows the HashMap to store duplicate Node keys.
        // This makes it necessary to rename a Node by first removing it
        // from the HashMap, then changing its nick, before finally putting
        // it back into the HashMap.
        // Likewise, the hashCode method for Edge depends on the hashCode
        // of its two Nodes, so all Edges emanating from the Node must
        // be removed from the HashMap before the Node is renamed. These must
        // then be put back into the HashMap after the Node has been renamed.
        
        // Create a List of Edges that will be affected by renaming oldNode.
        Iterator edgeIt = _edges.keySet().iterator();
        LinkedList changedEdges = new LinkedList();
        while (edgeIt.hasNext()) {
            Edge edge = (Edge) edgeIt.next();
            if (edge.getSource().equals(oldNode) || edge.getTarget().equals(oldNode)) {
                changedEdges.add(edge);
            }
        }
        
        // Remove all affected edges from the HashMap.
        Iterator changedIt = changedEdges.iterator();
        while (changedIt.hasNext()) {
            Edge edge = (Edge) changedIt.next();
            _edges.remove(edge);
        }
        
        // Remove, rename and replace the oldNode in the HashMap.
        oldNode = get(oldNode);
        _nodes.remove(oldNode);
        oldNode.setNick(nick);
        _nodes.put(oldNode, oldNode);
        
        // Put all the affected edges back into the HashMap.
        changedIt = changedEdges.iterator();
        while (changedIt.hasNext()) {
            Edge edge = (Edge) changedIt.next();
            _edges.put(edge, edge);
        }
        
        if (getConnectedNodes().contains(oldNode)) {
            // The changed node is in the graph, so it needs redrawing.
            if (autoSaveEachFrame()) makeNextImage();
        }
        
    }
    
    // Return true if the Graph contains the Node.
    // (This does not necessarily imply that the Node is visible).
    public boolean contains(Node node) {
        return _nodes.containsKey(node);
    }
    
    // Return true if the Graph contains the Edge.
    public boolean contains(Edge edge) {
        return _edges.containsKey(edge);
    }
    
    // Return the Graph's Node that has the same nick as the supplied Node.
    public Node get(Node node) {
        return (Node) _nodes.get(node);
    }
    
    // Return the Graph's Edge that matched the supplied Edge.
    public Edge get(Edge edge) {
        return (Edge) _edges.get(edge);
    }
    
    public String toString() {
        return "Graph: " + _nodes.size() + " nodes and " + _edges.size() + " edges.";
    }

    public String toString2() {
        return "Nodes:\n" + _nodes + "\nEdges:\n" + _edges;
    }
    
    // Apply the temporal decay to the Graph.
    public void decay(double amount) {
        
        // Decrease all Edge weightings and remove non-positive Edges.
        Iterator edgeIt = _edges.keySet().iterator();
        while (edgeIt.hasNext()) {
            Edge edge = (Edge) edgeIt.next();
            edge.setWeight(edge.getWeight() - amount);
            if (edge.getWeight() <= 0) {
                edgeIt.remove();
            }
        }
        
        // Decrease all Node weightings, but do not allow them to be negative.
        Iterator nodeIt = _nodes.keySet().iterator();
        while (nodeIt.hasNext()) {
            Node node = (Node) nodeIt.next();
            node.setWeight(node.getWeight() - amount);
            if (node.getWeight() < 0) {
                node.setWeight(0);
            }
        }
        
    }
    
    // Returns the set of all Nodes that have emanating Edges.
    // This therefore returns all Nodes that will be visible in the drawing.
    private HashSet getConnectedNodes() {
        HashSet connectedNodes = new HashSet();
        Iterator edgeIt = _edges.keySet().iterator();
        while (edgeIt.hasNext()) {
            Edge edge = (Edge) edgeIt.next();
            connectedNodes.add(edge.getSource());
            connectedNodes.add(edge.getTarget());
        }
        return connectedNodes;
    }

    // Applies the spring embedder.
    public void doLayout(int iterations) {
        
        // For performance, copy each set into an array.
        HashSet visibleNodes = getConnectedNodes();
        Node[] nodes = (Node[]) visibleNodes.toArray(new Node[visibleNodes.size()]);
        Edge[] edges = (Edge[]) _edges.keySet().toArray(new Edge[_edges.size()]);
        
        double k = config.k;
        double c = config.c;
        // Repulsive forces between nodes that are further apart than this are ignored.
        double maxRepulsiveForceDistance = config.maxRepulsiveForceDistance;
                
        // For each iteration...
        for (int it = 0; it < iterations; it++) {
            
            // Calculate forces acting on nodes due to node-node repulsions...
            for (int a = 0; a < nodes.length; a++) {
                for (int b = a + 1; b < nodes.length; b++) {
                    Node nodeA = nodes[a];
                    Node nodeB = nodes[b];
                    
                    double deltaX = nodeB.getX() - nodeA.getX();
                    double deltaY = nodeB.getY() - nodeA.getY();
                    
                    double distanceSquared = deltaX * deltaX + deltaY * deltaY;
                    
                    if (distanceSquared < 0.01) {
                        deltaX = Math.random() / 10 + 0.1;
                        deltaY = Math.random() / 10 + 0.1;
                        distanceSquared = deltaX * deltaX + deltaY * deltaY;
                    }
                    
                    double distance = Math.sqrt(distanceSquared);
                    
                    if (distance < maxRepulsiveForceDistance) {
                        double repulsiveForce = (k * k / distance);
                        
                        nodeB.setFX(nodeB.getFX() + (repulsiveForce * deltaX / distance));
                        nodeB.setFY(nodeB.getFY() + (repulsiveForce * deltaY / distance));
                        nodeA.setFX(nodeA.getFX() - (repulsiveForce * deltaX / distance));
                        nodeA.setFY(nodeA.getFY() - (repulsiveForce * deltaY / distance));
                    }
                }
            }
            
            // Calculate forces acting on nodes due to edge attractions.
            for (int e = 0; e < edges.length; e++) {
                Edge edge = edges[e];
                Node nodeA = edge.getSource();
                Node nodeB = edge.getTarget();
                
                double deltaX = nodeB.getX() - nodeA.getX();
                double deltaY = nodeB.getY() - nodeA.getY();
                
                double distanceSquared = deltaX * deltaX + deltaY * deltaY;
    
                // Avoid division by zero error or Nodes flying off to
                // infinity.  Pretend there is an arbitrary distance between
                // the Nodes.
                if (distanceSquared < 0.01) {
                    deltaX = Math.random() / 10 + 0.1;
                    deltaY = Math.random() / 10 + 0.1;
                    distanceSquared = deltaX * deltaX + deltaY * deltaY;
                }
                
                double distance = Math.sqrt(distanceSquared);
                
                if (distance >  maxRepulsiveForceDistance) {
                    distance = maxRepulsiveForceDistance;
                }
                
                distanceSquared = distance * distance;
                
                double attractiveForce = (distanceSquared - k * k) / k;
                
                // Make edges stronger if people know each other.
                double weight = edge.getWeight();
                if (weight < 1) {
                    weight = 1;
                }
                attractiveForce *= (Math.log(weight) * 0.5) + 1;
            
                nodeB.setFX(nodeB.getFX() - attractiveForce * deltaX / distance);
                nodeB.setFY(nodeB.getFY() - attractiveForce * deltaY / distance);
                nodeA.setFX(nodeA.getFX() + attractiveForce * deltaX / distance);
                nodeA.setFY(nodeA.getFY() + attractiveForce * deltaY / distance);
                
            }
            
            // Now move each node to its new location...
            for (int a = 0; a < nodes.length; a++) {
                Node node = nodes[a];
                
                double xMovement = c * node.getFX();
                double yMovement = c * node.getFY();
                
                // Limit movement values to stop nodes flying into oblivion.
                double max = config.maxNodeMovement;
                if (xMovement > max) {
                    xMovement = max;
                }
                else if (xMovement < -max) {
                    xMovement = -max;
                }
                if (yMovement > max) {
                    yMovement = max;
                }
                else if (yMovement < -max) {
                    yMovement = -max;
                }
                
                node.setX(node.getX() + xMovement);
                node.setY(node.getY() + yMovement);
                
                // Reset the forces
                node.setFX(0);
                node.setFY(0);
            }
            
        }
        
    }

    // Work out the drawing boundaries...
    public void calcBounds(int width, int height) {
        
        minX = Double.POSITIVE_INFINITY;
        maxX = Double.NEGATIVE_INFINITY;
        minY = Double.POSITIVE_INFINITY;
        maxY = Double.NEGATIVE_INFINITY;
        maxWeight = 0;
        
        HashSet nodes = getConnectedNodes();
        Iterator nodeIt = nodes.iterator();
        while (nodeIt.hasNext()) {
            Node node = (Node) nodeIt.next();
            
            if (node.getX() > maxX) {
                maxX = node.getX();
            }
            if (node.getX() < minX) {
                minX = node.getX();
            }
            if (node.getY() > maxY) {
                maxY = node.getY();
            }
            if (node.getY() < minY) {
                minY = node.getY();
            }
        }
        
        // Increase size if too small.
        double minSize = config.minDiagramSize;
        if (maxX - minX < minSize) {
            double midX = (maxX + minX) / 2;
            minX = midX - (minSize / 2);
            maxX = midX + (minSize / 2);
        }
        if (maxY - minY < minSize) {
            double midY = (maxY + minY) / 2;
            minY = midY - (minSize / 2);
            maxY = midY + (minSize / 2);
        }
        
        // Work out the maximum weight.
        Iterator edgeIt = _edges.keySet().iterator();
        while (edgeIt.hasNext()) {
            Edge edge = (Edge) edgeIt.next();
            if (edge.getWeight() > maxWeight) {
                maxWeight = edge.getWeight();
            }
        }
        
        // Jibble the boundaries to maintain the aspect ratio.
        double xyRatio = ((maxX - minX) / (maxY - minY)) / (width / height);
        if (xyRatio > 1) {
            // diagram is wider than it is high.
            double dy = maxY - minY;
            dy = dy * xyRatio - dy;
            minY = minY - dy / 2;
            maxY = maxY + dy / 2;
        }
        else if (xyRatio < 1) {
            // Diagram is higher than it is wide.
            double dx = maxX - minX;
            dx = dx / xyRatio - dx;
            minX = minX - dx / 2;
            maxX = maxX + dx / 2;
        }
        
    }

    public BufferedImage drawImage(int width, int height, int borderSize, int nodeRadius, double edgeThreshold, boolean showEdges) {

	int numcols = width / 100;
	int colwidth = width / numcols;
	int numrows = height / 100;
	int rowheight = height / numrows;

        HashSet nodes = getConnectedNodes();
       
        // Now actually draw the thing...

	int realWidth = width;
	int realHeight = height;

        BufferedImage image = new BufferedImage(width, height, BufferedImage.TYPE_INT_ARGB);
        Graphics2D g = image.createGraphics();

        g.setColor(config.backgroundColor);
        g.fillRect(1, 1, width - 2, height - 2);
        
        g.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);

        g.setColor(config.borderColor);
        g.drawRect(0, 0, width - 1, height - 1);
        
        width = width - borderSize * 3;  // note the 3 (gives more border on right side)
        height = height - borderSize * 2;

	boolean debug = true;
	ArrayList boxes = new ArrayList();
	Map rectangleNames = new HashMap();
	Map nameRectangles = new TreeMap();
	Map nodeRectangles = new TreeMap();

	int x_count = 0;
	int y_count = 0;
	Color boxNameColor = new Color(200, 200, 200);
	Color boxNameColorBorder = new Color(100, 100, 100);
	Color boxBorderColor = new Color(200, 200, 255);
	Font boxNameFont = new Font("SansSerif", Font.PLAIN, 8);
	Font boxNameFontBorder = new Font("SansSerif", Font.PLAIN, 36);
	for (int x=0; x < width+colwidth; x += colwidth) {
	    for (int y=0; y < height+rowheight; y += rowheight) {
		Rectangle r = new Rectangle(x, y, colwidth, rowheight);
		boxes.add(r);
		String name = new StringBuffer().append(box_alpha[x_count]).append("-").append(box_alpha[y_count]).toString();
		rectangleNames.put(r, name);
		nameRectangles.put(name, r);

		if (debug) System.out.println("Rectangle " + name + ": " + r);
		g.setColor(boxBorderColor);
		g.draw(r);
		String label = name;
		if (y == 0 || x == 0) {
		    g.setColor(boxNameColorBorder);
		    g.setFont(boxNameFontBorder);
		    if (y == 0 && x > 0) {
			label = box_alpha[x_count];
		    } else if (x == 0 && y > 0) {
			label = box_alpha[y_count];
		    } else {
			label = box_alpha[0];
		    }
		} else {
		    g.setColor(boxNameColor);
		    g.setFont(boxNameFont);
		}
		FontMetrics metrics = g.getFontMetrics(g.getFont());
		g.drawString(label, r.x + 3, r.y + metrics.getHeight()+3);
		y_count++;
	    }
	    y_count = 0;
	    x_count++;
	}

        g.setColor(config.channelColor);
        g.setFont(new Font("SansSerif", Font.BOLD, 64));
        g.drawString(_label, borderSize + 20, 80);

        g.setColor(config.titleColor);
        g.setFont(new Font("SansSerif", Font.BOLD, 18));
        g.drawString(title, borderSize, borderSize - nodeRadius - 15);
        g.drawString(_caption, borderSize, height + borderSize * 2 - 5 - 50);
        g.setFont(new Font("SansSerif", Font.PLAIN, 12));
	String _desc1 = desc1;
	if (!config.getBoolean("CustomDescription")) {
	    _desc1 = "Generated by " + config.nick + " on " + config.server + " using " + SocialNetworkBot.VERSION;
	}
        g.drawString(_desc1, borderSize, height + borderSize * 2 - 5 - 30);
        g.drawString(desc2, borderSize, height + borderSize * 2 - 5 - 15);
	if (desc3 == null) {
	    g.drawString("http://www.jibble.org/piespy/ - This frame was drawn at " + new Date(), borderSize, height + borderSize * 2 - 5);
	} else {
	    g.drawString(desc3, borderSize, height + borderSize * 2 - 5);
	}
        
        // Draw all edges...
        Iterator edgeIt = _edges.keySet().iterator();
        while (edgeIt.hasNext()) {
            Edge edge = (Edge) edgeIt.next();
            
            if (edge.getWeight() < edgeThreshold) {
                continue;
            }
            
            double weight = edge.getWeight();
            
            Node nodeA = edge.getSource();
            Node nodeB = edge.getTarget();
            int x1 = (int) (width * (nodeA.getX() - minX) / (maxX - minX)) + borderSize;
            int y1 = (int) (height * (nodeA.getY() - minY) / (maxY - minY)) + borderSize;
            int x2 = (int) (width * (nodeB.getX() - minX) / (maxX - minX)) + borderSize;
            int y2 = (int) (height * (nodeB.getY() - minY) / (maxY - minY)) + borderSize;
            g.setStroke(new BasicStroke((float) (Math.log(weight + 1) * 0.5) + 1));
            int alpha = 102 + (int) (153 * weight / maxWeight);
            g.setColor(new Color(config.edgeColor.getRed(), config.edgeColor.getGreen(), config.edgeColor.getBlue(), alpha));
            if (showEdges) {
                g.drawLine(x1, y1, x2, y2);
            }
        }
        
        // Draw all nodes...
        g.setStroke(new BasicStroke(2.0f));
        g.setFont(new Font("SansSerif", Font.PLAIN, 10));
        Iterator nodeIt = nodes.iterator();
        while (nodeIt.hasNext()) {
            Node node = (Node) nodeIt.next();
            int x1 = (int) (width * (node.getX() - minX) / (maxX - minX)) + borderSize;
            int y1 = (int) (height * (node.getY() - minY) / (maxY - minY)) + borderSize;
            //int newNodeRadius = (int) Math.log((node.getWeight() + 1) / 10) + nodeRadius;
            g.setColor(config.nodeColor);
            g.fillOval(x1 - nodeRadius, y1 - nodeRadius, nodeRadius * 2, nodeRadius * 2);
            g.setColor(config.edgeColor);
            g.drawOval(x1 - nodeRadius, y1 - nodeRadius, nodeRadius * 2, nodeRadius * 2);
            //g.setColor(Color.white);
            //g.drawString(node.toString(), x1 + nodeRadius + 1, y1 - nodeRadius);
            //g.drawString(node.toString(), x1 + nodeRadius - 1, y1 - nodeRadius);
            //g.drawString(node.toString(), x1 + nodeRadius, y1 - nodeRadius + 1);
            //g.drawString(node.toString(), x1 + nodeRadius, y1 - nodeRadius - 1);
            g.setColor(config.labelColor);
            g.drawString(node.toString(), x1 + nodeRadius, y1 - nodeRadius);

	    for (Iterator i = boxes.iterator(); i.hasNext();) {
		Rectangle r = (Rectangle) i.next();
		if (r.contains(x1, y1)) {
		    System.out.println("Placing " + node.toString() + " <" + x1 + "," + y1 + "> in " + r);
		    nodeRectangles.put(node, r);
		}
	    }

        }

	for (Iterator i = nodeRectangles.entrySet().iterator();i.hasNext();) {
	    Map.Entry entry = (Map.Entry) i.next();
	    System.out.println(entry.getKey() + ": " + rectangleNames.get(entry.getValue()));
	}
        
	boolean createIndex = config.getBoolean("AttachNodeIndex");
	if (createIndex) {
	    Color indexColor = Color.BLACK;
	    Font indexFont = new Font("SansSerif", Font.PLAIN, 8);
	    FontMetrics fontMetrics = g.getFontMetrics(indexFont);
	    int indexRowHeight = fontMetrics.getHeight()+4;
	    
	    int nodeCount = nodes.size();
	    int indexColumnWidth = 200;
	    int maxIndexColumns = realWidth / indexColumnWidth;
	    int indexRowsPerColumn = config.getInt("NodeIndexHeight");
	    int indexImageHeight = indexRowsPerColumn * indexRowHeight + indexRowHeight*2;

	    BufferedImage indexImage = new BufferedImage(realWidth, indexImageHeight, BufferedImage.TYPE_INT_ARGB);
	    Graphics2D ig = indexImage.createGraphics();

	    ig.setColor(config.backgroundColor);
	    ig.fillRect(0, 0, realWidth, indexImageHeight);
	    ig.setColor(indexColor);
	    ig.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
	    
	    int rowCount = 0;
	    int columnCount = 0;
	    int maxY = 0;
	    Color pyjamasColor = new Color(220, 220, 255);
	    boolean pyjamas = false;
	    for (Iterator i = nodeRectangles.entrySet().iterator();i.hasNext();) {
		Map.Entry entry = (Map.Entry) i.next();
		String nick = ((Node) entry.getKey()).toString();
		String boxname = (String) rectangleNames.get(entry.getValue());

		int rowx = columnCount*indexColumnWidth+5;
		rowx += 3*columnCount;

		Rectangle rowRect = new Rectangle(rowx, rowCount*indexRowHeight+5,
						  indexColumnWidth, indexRowHeight);
		if (pyjamas) {
		    ig.setColor(pyjamasColor);
		    ig.fillRect(rowRect.x, rowRect.y+3, rowRect.width, rowRect.height);
		}
		int boxnamewidth = fontMetrics.stringWidth(boxname);
		System.out.println("name " + boxname + " width " + boxnamewidth);
		System.out.println(rowRect.toString());
		int boxnameX = rowRect.x + rowRect.width - (int) (((float) boxnamewidth)*1.5F);
		boxnameX -= 3;

		ig.setColor(indexColor);
		ig.drawString(nick, rowRect.x, rowRect.y+rowRect.height);
		ig.drawString(boxname, boxnameX, rowRect.y+rowRect.height);

		if (rowRect.y+rowRect.height > maxY) maxY = rowRect.y+rowRect.height;

		pyjamas = !pyjamas;
		rowCount++;
		if (rowCount > indexRowsPerColumn) {
		    columnCount++;
		    rowCount = 0;
		    pyjamas = false;
		    if (columnCount > maxIndexColumns) {
			System.out.println("Warning: too many columns (" + columnCount + ")");
		    }
		}
	    }

	    BufferedImage combImage = new BufferedImage(realWidth, realHeight+maxY+indexRowHeight*2, BufferedImage.TYPE_INT_ARGB);
	    Graphics2D cf = combImage.createGraphics();
	    cf.drawImage(image, null, 0, 0);
	    cf.drawImage(indexImage, null, 0, realHeight);
	    image = combImage;
	    
	}

        return image;
    }
    
    public int getFrameCount() {
        return _frameCount;
    }
    
    public String getLabel() {
        return _label;
    }
    
    public void setCaption(String caption) {
        _caption = caption;
    }
    
    public void setLastFile(File file) {
        _lastFile = file;
    }
    
    public File getLastFile() {
        return _lastFile;
    }
    
    public void makeNextImage() {
        _frameCount++;
        String strippedChannel = _label.toLowerCase().substring(1);
        
        File dir = new File(config.outputDirectory, strippedChannel);
        dir.mkdir();
            
        doLayout(config.springEmbedderIterations);
        calcBounds(config.outputWidth, config.outputHeight);
            
        try {
            BufferedImage image = drawImage(config.outputWidth, config.outputHeight, config.borderSize, config.nodeRadius, config.edgeThreshold, config.showEdges);

            // Write the archive image.
            File file = new File(dir, strippedChannel + "-" + _nf.format(_frameCount) + ".png");
            if (config.createArchive) {
                ImageIO.write(image, "png", file);
                _lastFile = file;
            }

            // Also save an image as channel-current.png.
            File current = new File(dir, strippedChannel + "-current.png");
            if (config.createCurrent) {
                ImageIO.write(image, "png", current);
                if (!config.createArchive) {
                    _lastFile = file;
                }
            }
            
            // Also serialize the graph object for later retrieval.
            if (config.createRestorePoints) {
                writeGraph();
            }

        }
        catch (Exception e) {
            System.out.println("PieSpy has gone wibbly: " + e);
            e.printStackTrace();
        }
        
        // Apply the temporal decay after each frame is created.
        decay(config.temporalDecayAmount);
    }
    
    // Serialize this Graph and write it to a File.
    public void writeGraph() {
        try {
            String strippedChannel = _label.toLowerCase().substring(1);
            File dir = new File(config.outputDirectory, strippedChannel);
            File file = new File(dir, strippedChannel + "-restore.dat");
            ObjectOutputStream oos = new ObjectOutputStream(new FileOutputStream(file));
            oos.writeObject(SocialNetworkBot.VERSION);
            oos.writeObject(this);
            oos.flush();
            oos.close();
        }
        catch (Exception e) {
            // Do nothing?
        }
    }

    public int getNodeCount() {
	return _nodes.size();
    }
    
    private String _label;
    private String _caption = "";
    private HashMap _nodes = new HashMap();
    private HashMap _edges = new HashMap();
    
    private double minX = Double.POSITIVE_INFINITY;
    private double maxX = Double.NEGATIVE_INFINITY;
    private double minY = Double.POSITIVE_INFINITY;
    private double maxY = Double.NEGATIVE_INFINITY;
    private double maxWeight = 0;

    private Configuration config;
    private int _frameCount = 0;
    private File _lastFile = null;
    
    private ArrayList _heuristics = new ArrayList();
    
    private NumberFormat _nf;
}
